/* -*- C++ -*-
 *
 * This file is a part of LEMON, a generic C++ optimization library
 *
 * Copyright (C) 2003-2006
 * Egervary Jeno Kombinatorikus Optimalizalasi Kutatocsoport
 * (Egervary Research Group on Combinatorial Optimization, EGRES).
 *
 * Permission to use, modify and distribute this software is granted
 * provided that this copyright notice appears in all copies. For
 * precise terms see the accompanying LICENSE file.
 *
 * This software is provided "AS IS" with no warranty of any kind,
 * express or implied, and with no claim as to its suitability for any
 * purpose.
 *
 */

#ifndef LEMON_STATIC_MAP_H
#define LEMON_STATIC_MAP_H

#include <algorithm>
#include <iostream>

#include <lemon/utility.h>
#include <lemon/bits/map_extender.h>
#include <lemon/bits/alteration_notifier.h>
#include <lemon/error.h>
#include <lemon/concept_check.h>
#include <lemon/concept/maps.h>

/// \ingroup graphmaps
///
///\file
///\brief Static sized graph maps.

namespace lemon {

  /// \ingroup graphmaps
  ///
  /// \brief Graph map with static sized storage.
  ///
  /// The StaticMap template class is graph map structure what
  /// does not indate automatically the map when a key is added to or 
  /// erased from the map rather it throws an exception. This map factory 
  /// uses the allocators to implement the container functionality.
  ///
  /// \param Registry The AlterationNotifier that will notify this map.
  /// \param Item The item type of the graph items.
  /// \param Value The value type of the map.
  /// 
  /// \author Balazs Dezso
  	

  template <typename _Graph, typename _Item, typename _Value>
  class StaticMap : public AlterationNotifier<_Item>::ObserverBase {
  public:

    /// \brief Exception class for unsinported exceptions.
    class UnsinportedOperation : public lemon::LogicError {
    public:
      virtual const char* exceptionName() const {
	return "lemon::StaticMap::UnsinportedOperation";
      }
    };

  private:
		
    typedef std::vector<_Value> Container;	

  public:

    /// The graph type of the map. 
    typedef _Graph Graph;
    /// The reference map tag.
    typedef True ReferenceMapTag;

    /// The key type of the map.
    typedef _Item Key;
    /// The value type of the map.
    typedef _Value Value;
    /// The const reference type of the map.
    typedef typename Container::const_reference ConstReference;
    /// The reference type of the map.
    typedef typename Container::reference Reference;

    typedef const Value ConstValue;
    typedef Value* Pointer;
    typedef const Value* ConstPointer;

    typedef AlterationNotifier<_Item> Registry;

    /// The map type.
    typedef StaticMap Map;
    /// The base class of the map.
    typedef typename Registry::ObserverBase Parent;

    /// \brief Constructor to attach the new map into the registry.
    ///
    /// It constructs a map and attachs it into the registry.
    /// It adds all the items of the graph to the map.
    StaticMap(const Graph& _g) : graph(&_g) {
      attach(_g.getNotifier(_Item()));
      build();
    }

    /// \brief Constructor uses given value to initialize the map. 
    ///
    /// It constructs a map uses a given value to initialize the map. 
    /// It adds all the items of the graph to the map.     
    StaticMap(const Graph& _g, const Value& _v) : graph(&_g) { 
      attach(_g.getNotifier(_Item()));
      unsigned size = graph->maxId(_Item()) + 1;
      container.reserve(size);
      container.resize(size, _v);
    }

    /// \brief Copy constructor
    ///
    /// Copy constructor.
    StaticMap(const StaticMap& _copy) : Parent(), graph(_copy.getGraph()) {
      if (_copy.attached()) {
	attach(*_copy.getRegistry());
	container = _copy.container;
      }
    }

    /// \brief Destrcutor
    ///
    /// Destructor.
    virtual ~StaticMap() {
      clear();
      if (attached()) {
	detach();
      }
    }


  private:

    StaticMap& operator=(const StaticMap&);

  protected:

    using Parent::attach;
    using Parent::detach;
    using Parent::attached;

    const Graph* getGraph() const {
      return graph;
    }

  public:

    /// \brief The subcript operator.
    ///
    /// The subscript operator. The map can be subscripted by the
    /// actual items of the graph. 
     
    Reference operator[](const Key& key) {
      return container[graph->id(key)];
    } 
		
    /// \brief The const subcript operator.
    ///
    /// The const subscript operator. The map can be subscripted by the
    /// actual items of the graph. 
     
    ConstReference operator[](const Key& key) const {
      return container[graph->id(key)];
    }


    /// \brief The setter function of the map.
    ///
    /// It the same as operator[](key) = value expression.
    ///
    void set(const Key& key, const Value& value) {
      (*this)[key] = value;
    }

  protected:

    /// \brief Adds a new key to the map.
    ///		
    /// It adds a new key to the map. It called by the observer registry
    /// and it overrides the add() member function of the observer base.
     
    void add(const Key&) {
      throw UnsinportedOperation();
    }

    /// \brief Erases a key from the map.
    ///
    /// Erase a key from the map. It called by the observer registry
    /// and it overrides the erase() member function of the observer base.     
    void erase(const Key&) {
      throw UnsinportedOperation();
    }

    /// Buildes the map.
		
    /// It buildes the map. It called by the observer registry
    /// and it overrides the build() member function of the observer base.

    void build() { 
      int size = graph->maxId(_Item()) + 1;
      container.reserve(size);
      container.resize(size);
    }

    /// Clear the map.

    /// It erase all items from the map. It called by the observer registry
    /// and it overrides the clear() member function of the observer base.     
    void clear() { 
      container.clear();
    }
    
  private:
		
    Container container;
    const Graph *graph;

  };

  /// \e
  template <typename _Base> 
  class StaticMappableGraphExtender : public _Base {
  public:

    typedef StaticMappableGraphExtender<_Base> Graph;
    typedef _Base Parent;

    typedef typename Parent::Node Node;
    typedef typename Parent::NodeIt NodeIt;

    typedef typename Parent::Edge Edge;
    typedef typename Parent::EdgeIt EdgeIt;

    
    template <typename _Value>
    class NodeMap 
      : public IterableMapExtender<StaticMap<Graph, Node, _Value> > {
    public:
      typedef StaticMappableGraphExtender Graph;
      typedef IterableMapExtender<StaticMap<Graph, Node, _Value> > Parent;

      NodeMap(const Graph& _g) 
	: Parent(_g) {}
      NodeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}

      NodeMap& operator=(const NodeMap& cmap) {
	return operator=<NodeMap>(cmap);
      }


      /// \brief Template assign operator.
      ///
      /// The given parameter should be conform to the ReadMap
      /// concecpt and could be indiced by the current item set of
      /// the NodeMap. In this case the value for each item
      /// is assigned by the value of the given ReadMap. 
      template <typename CMap>
      NodeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Node, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Node it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }

    };

    template <typename _Value>
    class EdgeMap 
      : public IterableMapExtender<StaticMap<Graph, Edge, _Value> > {
    public:
      typedef StaticMappableGraphExtender Graph;
      typedef IterableMapExtender<StaticMap<Graph, Edge, _Value> > Parent;

      EdgeMap(const Graph& _g) 
	: Parent(_g) {}
      EdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}

      EdgeMap& operator=(const EdgeMap& cmap) {
	return operator=<EdgeMap>(cmap);
      }

      template <typename CMap>
      EdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Edge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Edge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };
    
  };

  /// \e
  template <typename _Base> 
  class StaticMappableUGraphExtender : 
    public StaticMappableGraphExtender<_Base> {
  public:

    typedef StaticMappableUGraphExtender Graph;
    typedef StaticMappableGraphExtender<_Base> Parent;

    typedef typename Parent::UEdge UEdge;

    template <typename _Value>
    class UEdgeMap 
      : public IterableMapExtender<StaticMap<Graph, UEdge, _Value> > {
    public:
      typedef StaticMappableUGraphExtender Graph;
      typedef IterableMapExtender<
	StaticMap<Graph, UEdge, _Value> > Parent;

      UEdgeMap(const Graph& _g) 
	: Parent(_g) {}
      UEdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}

      UEdgeMap& operator=(const UEdgeMap& cmap) {
	return operator=<UEdgeMap>(cmap);
      }

      template <typename CMap>
      UEdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<UEdge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	UEdge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };

  };

  template <typename _Base>
  class StaticMappableBpUGraphExtender : public _Base {
  public:

    typedef _Base Parent;
    typedef StaticMappableBpUGraphExtender Graph;

    typedef typename Parent::Node Node;
    typedef typename Parent::ANode ANode;
    typedef typename Parent::BNode BNode;
    typedef typename Parent::Edge Edge;
    typedef typename Parent::UEdge UEdge;
    
    template <typename _Value>
    class ANodeMap 
      : public IterableMapExtender<StaticMap<Graph, ANode, _Value> > {
    public:
      typedef StaticMappableBpUGraphExtender Graph;
      typedef IterableMapExtender<StaticMap<Graph, ANode, _Value> > 
      Parent;
    
      ANodeMap(const Graph& _g) 
	: Parent(_g) {}
      ANodeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      ANodeMap& operator=(const ANodeMap& cmap) {
	return operator=<ANodeMap>(cmap);
      }
    

      /// \brief Template assign operator.
      ///
      /// The given parameter should be conform to the ReadMap
      /// concept and could be indiced by the current item set of
      /// the ANodeMap. In this case the value for each item
      /// is assigned by the value of the given ReadMap. 
      template <typename CMap>
      ANodeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<ANode, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	ANode it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    
    };

    template <typename _Value>
    class BNodeMap 
      : public IterableMapExtender<StaticMap<Graph, BNode, _Value> > {
    public:
      typedef StaticMappableBpUGraphExtender Graph;
      typedef IterableMapExtender<StaticMap<Graph, BNode, _Value> > 
      Parent;
    
      BNodeMap(const Graph& _g) 
	: Parent(_g) {}
      BNodeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      BNodeMap& operator=(const BNodeMap& cmap) {
	return operator=<BNodeMap>(cmap);
      }
    

      /// \brief Template assign operator.
      ///
      /// The given parameter should be conform to the ReadMap
      /// concept and could be indiced by the current item set of
      /// the BNodeMap. In this case the value for each item
      /// is assigned by the value of the given ReadMap. 
      template <typename CMap>
      BNodeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<BNode, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	BNode it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    
    };

  protected:

    template <typename _Value>
    class NodeMapBase : public Parent::NodeNotifier::ObserverBase {
    public:
      typedef StaticMappableBpUGraphExtender Graph;

      typedef Node Key;
      typedef _Value Value;

      /// The reference type of the map;
      typedef typename BNodeMap<_Value>::Reference Reference;
      /// The pointer type of the map;
      typedef typename BNodeMap<_Value>::Pointer Pointer;
      
      /// The const value type of the map.
      typedef const Value ConstValue;
      /// The const reference type of the map;
      typedef typename BNodeMap<_Value>::ConstReference ConstReference;
      /// The pointer type of the map;
      typedef typename BNodeMap<_Value>::ConstPointer ConstPointer;

      typedef True ReferenceMapTag;

      NodeMapBase(const Graph& _g) 
	: graph(&_g), bNodeMap(_g), aNodeMap(_g) {
	Parent::NodeNotifier::ObserverBase::attach(_g.getNotifier(Node()));
      }
      NodeMapBase(const Graph& _g, const _Value& _v) 
	: graph(&_g), bNodeMap(_g, _v), 
	  aNodeMap(_g, _v) {
	Parent::NodeNotifier::ObserverBase::attach(_g.getNotifier(Node()));
      }

      virtual ~NodeMapBase() {      
	if (Parent::NodeNotifier::ObserverBase::attached()) {
	  Parent::NodeNotifier::ObserverBase::detach();
	}
      }
    
      ConstReference operator[](const Key& node) const {
	if (Parent::aNode(node)) {
	  return aNodeMap[node];
	} else {
	  return bNodeMap[node];
	}
      } 

      Reference operator[](const Key& node) {
	if (Parent::aNode(node)) {
	  return aNodeMap[node];
	} else {
	  return bNodeMap[node];
	}
      }

      void set(const Key& node, const Value& value) {
	if (Parent::aNode(node)) {
	  aNodeMap.set(node, value);
	} else {
	  bNodeMap.set(node, value);
	}
      }

    protected:
      
      virtual void add(const Node&) {}
      virtual void add(const std::vector<Node>&) {}
      virtual void erase(const Node&) {}
      virtual void erase(const std::vector<Node>&) {}
      virtual void clear() {}
      virtual void build() {}

      const Graph* getGraph() const { return graph; }
      
    private:
      const Graph* graph;
      BNodeMap<_Value> bNodeMap;
      ANodeMap<_Value> aNodeMap;
    };
    
  public:

    template <typename _Value>
    class NodeMap 
      : public IterableMapExtender<NodeMapBase<_Value> > {
    public:
      typedef StaticMappableBpUGraphExtender Graph;
      typedef IterableMapExtender< NodeMapBase<_Value> > Parent;
    
      NodeMap(const Graph& _g) 
	: Parent(_g) {}
      NodeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      NodeMap& operator=(const NodeMap& cmap) {
	return operator=<NodeMap>(cmap);
      }
    

      /// \brief Template assign operator.
      ///
      /// The given parameter should be conform to the ReadMap
      /// concept and could be indiced by the current item set of
      /// the NodeMap. In this case the value for each item
      /// is assigned by the value of the given ReadMap. 
      template <typename CMap>
      NodeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Node, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Node it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    
    };



    template <typename _Value>
    class EdgeMap 
      : public IterableMapExtender<StaticMap<Graph, Edge, _Value> > {
    public:
      typedef StaticMappableBpUGraphExtender Graph;
      typedef IterableMapExtender<StaticMap<Graph, Edge, _Value> > Parent;
    
      EdgeMap(const Graph& _g) 
	: Parent(_g) {}
      EdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      EdgeMap& operator=(const EdgeMap& cmap) {
	return operator=<EdgeMap>(cmap);
      }
    
      template <typename CMap>
      EdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Edge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Edge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };

    template <typename _Value>
    class UEdgeMap 
      : public IterableMapExtender<StaticMap<Graph, UEdge, _Value> > {
    public:
      typedef StaticMappableBpUGraphExtender Graph;
      typedef IterableMapExtender<StaticMap<Graph, UEdge, _Value> > 
      Parent;
    
      UEdgeMap(const Graph& _g) 
	: Parent(_g) {}
      UEdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      UEdgeMap& operator=(const UEdgeMap& cmap) {
	return operator=<UEdgeMap>(cmap);
      }
    
      template <typename CMap>
      UEdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<UEdge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	UEdge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };
  
  };

}

#endif
